//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Import/SpecularDataImportWidget.cpp
//! @brief     Implements class SpecularDataImportWidget
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Import/SpecularDataImportWidget.h"
#include "GUI/Model/Data/SpecularDataItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/Support/IO/AbstractDataLoaderResultModel.h"
#include "GUI/View/Loaders/DataLoaderUtil.h"
#include "GUI/View/Loaders/QREDataLoader.h"
#include "GUI/View/PlotSpecular/SpecularDataCanvas.h"
#include "GUI/View/Widget/StyledToolbar.h"
#include <QAction>

SpecularDataImportWidget::SpecularDataImportWidget(QWidget* parent)
    : DataAccessWidget(parent)
    , m_loader(nullptr)
{
    setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    setGeometry(0, 0, 861, 580);

    auto* vbox = new QVBoxLayout;
    setLayout(vbox);

    auto* splitter = new QSplitter(Qt::Horizontal);
    vbox->addWidget(splitter);
    splitter->setChildrenCollapsible(false);

    //... Left side of splitter: import filter settings

    auto* importer_widget = new QWidget;
    splitter->addWidget(importer_widget);

    auto* importer_vbox = new QVBoxLayout;
    importer_widget->setLayout(importer_vbox);
    importer_vbox->setSpacing(20);

    auto* import_settings_widget = new QWidget;
    importer_vbox->addWidget(import_settings_widget);
    import_settings_widget->setWindowTitle("Import settings");

    auto* import_settings_vbox = new QVBoxLayout;
    import_settings_widget->setLayout(import_settings_vbox);

    m_properties_widget = new QWidget; // to be populated later
    import_settings_vbox->addWidget(m_properties_widget);

    m_warn_widget = new QWidget;
    import_settings_vbox->addWidget(m_warn_widget);

    auto* warn_vbox = new QVBoxLayout;
    m_warn_widget->setLayout(warn_vbox);

    auto* warn_row = new QHBoxLayout;
    warn_vbox->addLayout(warn_row);

    auto* warn_icon = new QLabel;
    warn_row->addWidget(warn_icon);
    warn_icon->setFixedSize(20, 20);
    warn_icon->setPixmap(QPixmap(":/images/warning_16x16.png"));

    auto* warn_label = new QLabel;
    warn_row->addWidget(warn_label);
    warn_label->setText("The file import issued warnings:");

    m_warn_list = new QListWidget;
    import_settings_vbox->addWidget(m_warn_list);

    //... Right side of splitter: data view

    auto* tab = new QTabWidget;
    splitter->addWidget(tab);

    auto* data_widget = new QWidget;
    tab->addTab(data_widget, "Table");

    auto* data_vbox = new QVBoxLayout;
    data_widget->setLayout(data_vbox);

    m_check_raw = new QCheckBox("Show file content");
    data_vbox->addWidget(m_check_raw);
    m_check_raw->setChecked(false);

    m_check_col = new QCheckBox("Show raw columns");
    data_vbox->addWidget(m_check_col);
    m_check_col->setChecked(false);

    m_check_fin = new QCheckBox("Show extracted columns");
    data_vbox->addWidget(m_check_fin);
    m_check_fin->setChecked(true);

    m_data_table = new QTableView;
    data_vbox->addWidget(m_data_table);

    auto* graph_widget = new QWidget;
    tab->addTab(graph_widget, "Graph");

    auto* graph_vbox = new QVBoxLayout;
    graph_widget->setLayout(graph_vbox);

    auto* plotToolbar = new StyledToolbar;
    graph_vbox->addWidget(plotToolbar);

    m_canvas = new SpecularDataCanvas;
    graph_vbox->addWidget(m_canvas);

    //...

    updatePropertiesEdits();

    connect(m_check_raw, &QCheckBox::stateChanged, this, &SpecularDataImportWidget::updatePreview);

    connect(m_check_col, &QCheckBox::stateChanged, this, &SpecularDataImportWidget::updatePreview);

    connect(m_check_fin, &QCheckBox::stateChanged, this, &SpecularDataImportWidget::updatePreview);

    connect(m_canvas->customPlot(), &QCustomPlot::axisClick, this,
            &SpecularDataImportWidget::onPlotAxisClicked);

    m_canvas->enableDeprecatedOnMousePress(false); // we have an own handler

    plotToolbar->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);
    for (auto* action : m_canvas->actionList())
        plotToolbar->addAction(action);
}

void SpecularDataImportWidget::setJobOrRealItem(QObject* real_item)
{
    DataAccessWidget::setJobOrRealItem(real_item);
    ASSERT(realItem());
    m_canvas->setJobOrRealItem(real_item);

    m_loader = dynamic_cast<AbstractDataLoader1D*>(realItem()->dataLoader());
    ASSERT(m_loader); // only items which have a loader are allowed for this widget. Every other
                      // items do not support this widget

    updatePropertiesEdits();
    updatePreview();
    connect(m_loader, &AbstractDataLoader::importSettingsChanged, this,
            &SpecularDataImportWidget::onPropertiesChanged, Qt::UniqueConnection);
}

QList<QAction*> SpecularDataImportWidget::actionList()
{
    return {};
}

void SpecularDataImportWidget::onContextMenuRequest(const QPoint& point)
{
    QMenu menu;
    for (auto* action : actionList())
        menu.addAction(action);
    menu.exec(point);
}

void SpecularDataImportWidget::onPlotAxisClicked(QCPAxis* axis, QCPAxis::SelectablePart /*part*/,
                                                 QMouseEvent* event)
{
    if (event->button() == Qt::RightButton && axis->axisType() == QCPAxis::atLeft) {
        QMenu menu;

        auto* lin = new QAction("Linear");
        connect(lin, &QAction::triggered, [this]() { realSpecularDataItem()->setLog(false); });
        lin->setCheckable(true);
        lin->setChecked(!realSpecularDataItem()->isLog());

        auto* log = new QAction("Logarithmic");
        connect(log, &QAction::triggered, [this]() { realSpecularDataItem()->setLog(true); });
        log->setCheckable(true);
        log->setChecked(realSpecularDataItem()->isLog());

        auto* ag = new QActionGroup(&menu);
        ag->addAction(lin);
        ag->addAction(log);

        menu.addAction(lin);
        menu.addAction(log);

        menu.exec(event->globalPos());
    }
}

void SpecularDataImportWidget::updatePropertiesEdits()
{
    for (auto* child : m_properties_widget->children())
        delete child;

    if (m_properties_widget->layout())
        delete m_properties_widget->layout();

    if (m_loader)
        m_loader->populateImportSettingsWidget(m_properties_widget);

    const bool hasChildren = !m_properties_widget->children().empty();

    m_properties_widget->setVisible(hasChildren);
}

AbstractDataLoader* SpecularDataImportWidget::selectedLoader()
{
    return new QREDataLoader;
}

void SpecularDataImportWidget::updatePreview()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);

    if (m_loader) {
        auto* oldModel = m_data_table->selectionModel(); // sic!! according to Qt docu
                                                         // of QAbstractItemView::setModel
        auto* resultModel = m_loader->createResultModel();
        if (resultModel != nullptr) {
            const auto originalSections = resultModel->sectionsOfColumnType(
                AbstractDataLoaderResultModel::ColumnType::fileContent);

            const auto rawSections =
                resultModel->sectionsOfColumnType(AbstractDataLoaderResultModel::ColumnType::raw);

            const auto processedSections = resultModel->sectionsOfColumnType(
                AbstractDataLoaderResultModel::ColumnType::processed);

            QSignalBlocker b1(m_check_raw);
            QSignalBlocker b2(m_check_col);
            QSignalBlocker b3(m_check_fin);

            if (originalSections.isEmpty()) {
                m_check_raw->setChecked(false);
                m_check_raw->setEnabled(false);
            } else
                m_check_raw->setEnabled(true);

            if (rawSections.isEmpty()) {
                m_check_col->setChecked(false);
                m_check_col->setEnabled(false);
            } else
                m_check_col->setEnabled(true);

            if (processedSections.isEmpty()) {
                m_check_fin->setChecked(false);
                m_check_fin->setEnabled(false);
            } else
                m_check_fin->setEnabled(true);

            m_data_table->setModel(resultModel);
            auto* horHeader = m_data_table->horizontalHeader();

            for (int section : originalSections)
                horHeader->setSectionHidden(section, !m_check_raw->isChecked());

            for (int section : rawSections)
                horHeader->setSectionHidden(section, !m_check_col->isChecked());

            for (int section : processedSections)
                horHeader->setSectionHidden(section, !m_check_fin->isChecked());

            // if the result model has a line column, then do not show the vertical header view
            const bool hasLinesColumn =
                !resultModel->sectionsOfColumnType(AbstractDataLoaderResultModel::ColumnType::line)
                     .isEmpty();
            m_data_table->verticalHeader()->setHidden(hasLinesColumn);

            m_data_table->resizeColumnsToContents();
        } else
            m_data_table->setModel(nullptr);

        delete oldModel;
    }

    if (m_loader && m_loader->numErrors() > 0) {
        m_warn_list->clear();
        m_warn_widget->show();

        auto warnings = m_loader->lineUnrelatedErrors();
        const int nLineRelatedWarnings = m_loader->numErrors() - warnings.size();

        if (nLineRelatedWarnings == 1)
            warnings << "There is a warning about one input line. "
                        "See the \"Parser warnings\" in the rightmost column in the data tab.";
        else if (nLineRelatedWarnings > 1)
            warnings << QString(
                            "%1 warnings about specific input lines. "
                            "See the \"Parser warnings\" in the rightmost column in the data tab.")
                            .arg(nLineRelatedWarnings);

        if (warnings.size() > 1)
            for (auto& w : warnings)
                w.prepend("* ");

        for (auto& w : warnings)
            new QListWidgetItem(w, m_warn_list);
    } else {
        m_warn_widget->hide();
    }

    QApplication::restoreOverrideCursor();
}

void SpecularDataImportWidget::onPropertiesChanged()
{
    m_loader->applyImportSettings(); // #baimport: may be duplicate
    QApplication::setOverrideCursor(Qt::WaitCursor);
    m_loader->processContents();
    QApplication::restoreOverrideCursor();

    // If there is a linked instrument, any change in import settings can break the compatibility.
    // Therefore check the compatibility and break the link  if necessary
    ASSERT(gProjectDocument.has_value());

    const auto* linkedInstrument =
        gProjectDocument.value()->instrumentModel()->findInstrumentItemById(
            realItem()->instrumentId());

    if (linkedInstrument)
        if (!linkedInstrument->alignedWith(realItem()))
            realItem()->unlinkFromInstrument();

    updatePreview();
}

QString SpecularDataImportWidget::currentFileName() const
{
    return realItem()->nativeFileName();
}
